// OutdoorLighting.cpp : Defines the entry point for the console application.
//

#include "LocalTime.h"
#include "Location.h"
#include "Ephemeris.h"
#include "SolarSpectrum.h"
#include "LunarSpectrum.h"
#include "LuminanceMapper.h"
#include <stdio.h>
#include <math.h>
#include <map>

#define PI 3.14159265

// Radians to degrees
#define DEGREES(x)	((x) * (180.0 / PI))

// Convert watts/m2 to cd/m2
// wm2 * 683 lux / 1wm2 * 1 nit / 3.14 lux
#define NITS(irradiance) (irradiance * 683.0 / 3.14)

static const Matrix3 XYZ2RGB( 3.240479, -0.969256 ,  0.055648,
                              -1.53715,  1.875991,  -0.204043,
                              -0.49853,  0.041556,   1.057311 );

static std::map<int, double> twilightLuminance;

static void InitTwilightZenithLuminanceLookup()
{
    twilightLuminance[5] = 2200;
    twilightLuminance[4] = 1800;
    twilightLuminance[3] = 1400;
    twilightLuminance[2] = 1200;
    twilightLuminance[1] = 710;
    twilightLuminance[0] = 400;
    twilightLuminance[-1] = 190;
    twilightLuminance[-2] = 77;
    twilightLuminance[-3] = 28;
    twilightLuminance[-4] = 9.4;
    twilightLuminance[-5] = 2.9;
    twilightLuminance[-6] = 0.9;
    twilightLuminance[-7] = 0.3;
    twilightLuminance[-8] = 0.11;
    twilightLuminance[-9] = 0.047;
    twilightLuminance[-10] = 0.021;
    twilightLuminance[-11] = 0.0092;
    twilightLuminance[-12] = 0.0031;
    twilightLuminance[-13] = 0.0022;
    twilightLuminance[-14] = 0.0019;
    twilightLuminance[-15] = 0.0018;
    twilightLuminance[-16] = 0.0018;

    // Convert from asb to cd / m2
    std::map<int, double>::iterator it;
    for (it = twilightLuminance.begin(); it != twilightLuminance.end(); it++)
    {
        (*it).second = ((*it).second / PI);
    }
}

// Forces an RGB value to be within [0,1] and applies gamma correction
static void ApplyGamma(Vector3& v, double oneOverGamma)
{
    if (v.x < 0) v.x = 0;
    if (v.y < 0) v.y = 0;
    if (v.z < 0) v.z = 0;

    double max = v.x;
    if (v.y > max) max = v.y;
    if (v.z > max) max = v.z;
    if (max > 1.0)
    {
        v.x /= max;
        v.y /= max;
        v.z /= max;
    }

    v.x = pow(v.x, oneOverGamma);
    v.y = pow(v.y, oneOverGamma);
    v.z = pow(v.z, oneOverGamma);
}

// Computes the luminance of the moon based on the simualated phase and distance from Earth
static double MoonLuminance(const Ephemeris* ephemeris)
{
    double luminance = 0;

    if (ephemeris)
    {
        Vector3 moonPos = ephemeris->GetMoonPositionHorizon();
        moonPos.Normalize();
        double moonAngle = DEGREES(asin(moonPos.y));

        const double Esm = 1905.0; // W/m2
        const double C = 0.072;
        const double Rm = 1738.1 * 1000.0; // m
        double d = ephemeris->GetMoonDistanceKM() * 1000.0;

        double mPhase = ephemeris->GetMoonPhaseAngle();

        double ePhase = PI - mPhase;
        while (ePhase < 0) ePhase += 2.0 * PI;

        // Earthshine:
        double Eem = 0.19 * 0.5 *
                     (1.0 - sin(ePhase / 2.0) * tan(ePhase / 2.0) * log(1.0 / tan(ePhase / 4.0)));

        // Total moonlight:
        double Em = ((2.0 * C * Rm * Rm) / (3.0 * d * d)) *
                    (Eem + Esm *
                     (1.0 - sin(mPhase / 2.0) * tan(mPhase / 2.0) * log(1.0 / tan(mPhase / 4.0))));

        double nits = NITS(Em);

        if (moonAngle < 0) // Handle twilight effects from moon
        {
            nits = nits * exp(1.1247 * moonAngle);
        }

        luminance = nits;
    
    }

    return luminance;
}

// Computes the scattered and transmitted irradiance of the sun through the atmosphere, as XYZ colors.
static void ComputeSunlight(const Vector3& sunPos, const Ephemeris *ephemeris, double turbidity, 
                     double altitude, Vector3& solarDirectXYZ, Vector3& solarScatteredXYZ)
{
    SolarSpectrum solarSpectrum;

    double cosZenith = sunPos.y;

    if (cosZenith > 0)
    {
        Spectrum solarDirect, solarScattered;

        double zenithAngle = acos(cosZenith);
        
        solarSpectrum.ApplyAtmosphericTransmittance(zenithAngle, cosZenith, 
            turbidity, altitude, solarDirect, solarScattered);

        solarDirectXYZ = solarDirect.ToXYZ();
        solarScatteredXYZ = solarScattered.ToXYZ();
    }
    else
    {
        // In twilight conditions, we lookup luminance based on experimental results
        // on cloudless nights.
        double solarAltitude = DEGREES(asin(sunPos.y));

        int lower = (int)floor(solarAltitude);
        int higher = (int)ceil(solarAltitude);

        double alpha = solarAltitude - lower;

        double a = twilightLuminance[lower];
        double b = twilightLuminance[higher];

        // Blend light from sunset
        const double epsilon = 0.001;
        Spectrum solarDirect, solarScattered;
        double zenithAngle = PI * 0.5 - epsilon;
        
        solarSpectrum.ApplyAtmosphericTransmittance(zenithAngle, cos(zenithAngle), 
            turbidity, altitude, solarDirect, solarScattered);
        solarDirectXYZ = solarDirect.ToXYZ();
        solarScatteredXYZ = solarScattered.ToXYZ();

        double Y = (1 - alpha) * a + alpha * b; // luminance per lookup table
        double x = 0.2; double y = 0.2;
        double X = x * (Y / y);
        double Z = (1.0 - x - y) * (Y / y);

        alpha = -solarAltitude / 5.0;;
        if (alpha > 1.0) alpha = 1.0;
        if (alpha < 0) alpha = 0;
        
        solarDirectXYZ = Vector3(0,0,0) * alpha + solarDirectXYZ * (1.0 - alpha);
        solarScatteredXYZ = Vector3(X, Y, Z) * alpha + solarScatteredXYZ * (1.0 - alpha);
    }
}

// Computes the scattered and transmitted irradiance of the moon through the atmosphere, as XYZ colors.
static void ComputeMoonlight(const Vector3& moonPos, const Ephemeris *ephemeris, double turbidity, 
                      double altitude, Vector3& lunarDirectXYZ, Vector3& lunarScatteredXYZ)
{
    double moonLuminance = MoonLuminance(ephemeris);

    LunarSpectrum lunarSpectrum;
    lunarSpectrum *= moonLuminance;

    double cosZenith = moonPos.y;
    double zenith = acos(cosZenith);

    Spectrum moonSpectrumEarthDirect, moonSpectrumEarthScattered;
    lunarSpectrum.ApplyAtmosphericTransmittance(zenith, cosZenith, turbidity, altitude,
            moonSpectrumEarthDirect, moonSpectrumEarthScattered);

    lunarDirectXYZ = moonSpectrumEarthDirect.ToXYZ();
    lunarScatteredXYZ = moonSpectrumEarthScattered.ToXYZ();
}

int main(int argc, void* argv[])
{
    // Initialize our lookup table for twilight conditions
    InitTwilightZenithLuminanceLookup();

    // Set the conditions of the simulation and echo them to the console
    const double turbidity = 2.2;

    LocalTime localTime;
    localTime.SetYear(2010);
    localTime.SetMonth(8);
    localTime.SetDay(24);
    localTime.SetHour(21);

    Location location;
    location.SetLatitude(40.7);
    location.SetLongitude(5.4);
    location.SetAltitude(800.0);

    printf("Simulating %d/%d/%d at %d:%02d:%02d\nlatitude %3.2f longitude %3.2f altitude %5.2f\n\n",
        localTime.GetMonth(), localTime.GetDay(), localTime.GetYear(),
        localTime.GetHour(), localTime.GetMinutes(), (int)localTime.GetSeconds(),
        location.GetLatitude(), location.GetLongitude(), location.GetAltitude());

    // Calculate the direct and scattered radiance from the sun and the moon
    Ephemeris *ephemeris = new Ephemeris;
    ephemeris->Update(localTime, location);

    Vector3 sunPos = ephemeris->GetSunPositionHorizon();
    sunPos.Normalize();
    Vector3 moonPos = ephemeris->GetMoonPositionHorizon();
    moonPos.Normalize();

    printf("The sun is %5.2f degrees above the horizon.\n", DEGREES(asin(sunPos.y)));
    printf("The moon is %5.2f degrees above the horizon.\n", DEGREES(asin(moonPos.y)));

    Vector3 solarDirectXYZ, solarScatteredXYZ, lunarDirectXYZ, lunarScatteredXYZ;

    ComputeSunlight(sunPos, ephemeris, turbidity, location.GetAltitude(), solarDirectXYZ, solarScatteredXYZ);

    ComputeMoonlight(moonPos, ephemeris, turbidity, location.GetAltitude(), lunarDirectXYZ, lunarScatteredXYZ);

    double ambientLuminance = solarScatteredXYZ.y + lunarScatteredXYZ.y;

    printf("\nXYZ colors, prior to tone-mapping:\n");
    printf("Direct sunlight:     %.4f, %.4f, %.4f\n", solarDirectXYZ.x, solarDirectXYZ.y, solarDirectXYZ.z);
    printf("Scattered sunlight:  %.4f, %.4f, %.4f\n", solarScatteredXYZ.x, solarScatteredXYZ.y, solarScatteredXYZ.z);
    printf("Direct moonlight:    %.4f, %.4f, %.4f\n", lunarDirectXYZ.x, lunarDirectXYZ.y, lunarDirectXYZ.z);
    printf("Scattered moonlight: %.4f, %.4f, %.4f\n", lunarScatteredXYZ.x, lunarScatteredXYZ.y, lunarScatteredXYZ.z);

    // Tone-map these values to displayable values
    LuminanceMapper luminanceMapper;
    luminanceMapper.SetMaxDisplayLuminance(100.0);
    luminanceMapper.SetSceneLogAvg(ambientLuminance, ambientLuminance);
    luminanceMapper.DurandMapperXYZ(&solarDirectXYZ);
    luminanceMapper.DurandMapperXYZ(&solarScatteredXYZ);
    luminanceMapper.DurandMapperXYZ(&lunarDirectXYZ);
    luminanceMapper.DurandMapperXYZ(&lunarScatteredXYZ);

    printf("\nXYZ colors, after tone-mapping:\n");
    printf("Direct sunlight:     %.4f, %.4f, %.4f\n", solarDirectXYZ.x, solarDirectXYZ.y, solarDirectXYZ.z);
    printf("Scattered sunlight:  %.4f, %.4f, %.4f\n", solarScatteredXYZ.x, solarScatteredXYZ.y, solarScatteredXYZ.z);
    printf("Direct moonlight:    %.4f, %.4f, %.4f\n", lunarDirectXYZ.x, lunarDirectXYZ.y, lunarDirectXYZ.z);
    printf("Scattered moonlight: %.4f, %.4f, %.4f\n", lunarScatteredXYZ.x, lunarScatteredXYZ.y, lunarScatteredXYZ.z);

    // Convert them to RGB
    Vector3 solarDirectRGB = solarDirectXYZ * XYZ2RGB;
    Vector3 solarScatteredRGB = solarScatteredXYZ * XYZ2RGB;
    Vector3 lunarDirectRGB = lunarDirectXYZ * XYZ2RGB;
    Vector3 lunarScatteredRGB = lunarScatteredXYZ * XYZ2RGB;

    printf("\nRGB colors, prior to scaling and gamma correction:\n");
    printf("Direct sunlight:     %.4f, %.4f, %.4f\n", solarDirectRGB.x, solarDirectRGB.y, solarDirectRGB.z);
    printf("Scattered sunlight:  %.4f, %.4f, %.4f\n", solarScatteredRGB.x, solarScatteredRGB.y, solarScatteredRGB.z);
    printf("Direct moonlight:    %.4f, %.4f, %.4f\n", lunarDirectRGB.x, lunarDirectRGB.y, lunarDirectRGB.z);
    printf("Scattered moonlight: %.4f, %.4f, %.4f\n", lunarScatteredRGB.x, lunarScatteredRGB.y, lunarScatteredRGB.z);

    // Apply gamma correction
    const double oneOverGamma = 1.0 / 1.8;
    ApplyGamma(solarDirectRGB, oneOverGamma);
    ApplyGamma(solarScatteredRGB, oneOverGamma);
    ApplyGamma(lunarDirectRGB, oneOverGamma);
    ApplyGamma(lunarScatteredRGB, oneOverGamma);

    printf("\nFinal RGB colors:\n");
    printf("Direct sunlight:     %.4f, %.4f, %.4f\n", solarDirectRGB.x, solarDirectRGB.y, solarDirectRGB.z);
    printf("Scattered sunlight:  %.4f, %.4f, %.4f\n", solarScatteredRGB.x, solarScatteredRGB.y, solarScatteredRGB.z);
    printf("Direct moonlight:    %.4f, %.4f, %.4f\n", lunarDirectRGB.x, lunarDirectRGB.y, lunarDirectRGB.z);
    printf("Scattered moonlight: %.4f, %.4f, %.4f\n", lunarScatteredRGB.x, lunarScatteredRGB.y, lunarScatteredRGB.z);

    printf("\nPress return to exit.\n");
    getchar();

    return 0;
}

